@@ -5,9 +5,9 @@ import socket
 import struct
 import threading
 try:
-    from urllib import splitport
+    from urllib import splitport  # type: ignore
 except ImportError:
-    from urllib.parse import splitport
+    from urllib.parse import splitport  # type: ignore
 
 import zlib
 from io import BytesIO
@@ -25,6 +25,26 @@ logger = logging.getLogger(__name__)
 class Protocol(threading.local):
     """
     This class is used by Client class to communicate with server.
+
+    Reference https://github.com/memcached/memcached/wiki/BinaryProtocolRevamped ::
+
+        Header structure
+        Byte/     0       |       1       |       2       |       3       |
+           /              |               |               |               |
+          |0 1 2 3 4 5 6 7|0 1 2 3 4 5 6 7|0 1 2 3 4 5 6 7|0 1 2 3 4 5 6 7|
+          +---------------+---------------+---------------+---------------+
+         0| Magic         | Opcode        | Key length                    |
+          +---------------+---------------+---------------+---------------+
+         4| Extras length | Data type     | vbucket id                    |
+          +---------------+---------------+---------------+---------------+
+         8| Total body length                                             |
+          +---------------+---------------+---------------+---------------+
+        12| Opaque                                                        |
+          +---------------+---------------+---------------+---------------+
+        16| CAS                                                           |
+          |                                                               |
+          +---------------+---------------+---------------+---------------+
+          Total 24 bytes
     """
     HEADER_STRUCT = '!BBHBBHLLQ'
     HEADER_SIZE = 24
@@ -163,9 +183,7 @@ class Protocol(threading.local):
         Reads data from socket.
 
         :param size: Size in bytes to be read.
-        :type size: int
         :return: Data from socket
-        :rtype: six.string_types
         """
         value = b''
         while len(value) < size:
@@ -248,7 +266,7 @@ class Protocol(threading.local):
         if not self._username or not self._password:
             return False
 
-        logger.info('Authenticating as %s', self._username)
+        logger.debug('Authenticating as %s', self._username)
         self._send(struct.pack(self.HEADER_STRUCT,
                                self.MAGIC['request'],
                                self.COMMANDS['auth_negotiation']['command'],
@@ -269,7 +287,7 @@ class Protocol(threading.local):
 
         if b'PLAIN' not in methods:
             raise AuthenticationNotSupported('This module only supports '
-                                             'PLAIN auth for now.')
+                                             'PLAIN auth for now.', status)
 
         method = b'PLAIN'
         auth = '\x00%s\x00%s' % (self._username, self._password)
@@ -288,10 +306,10 @@ class Protocol(threading.local):
             return False
 
         if status == self.STATUS['auth_error']:
-            raise InvalidCredentials("Incorrect username or password")
+            raise InvalidCredentials("Incorrect username or password", status)
 
         if status != self.STATUS['success']:
-            raise MemcachedException('Code: %d Message: %s' % (status, extra_content))
+            raise MemcachedException('Code: %d Message: %s' % (status, extra_content), status)
 
         logger.debug('Auth OK. Code: %d Message: %s', status, extra_content)
 
@@ -419,12 +437,38 @@ class Protocol(threading.local):
             if status == self.STATUS['server_disconnected']:
                 return None, None
 
-            raise MemcachedException('Code: %d Message: %s' % (status, extra_content))
+            raise MemcachedException('Code: %d Message: %s' % (status, extra_content), status)
 
         flags, value = struct.unpack('!L%ds' % (bodylen - 4, ), extra_content)
 
         return self.deserialize(value, flags), cas
 
+    def noop(self):
+        """
+        Send a NOOP command
+
+        :return: Returns the status.
+        :rtype: int
+        """
+        logger.debug('Sending NOOP')
+        data = struct.pack(self.HEADER_STRUCT +
+                           self.COMMANDS['noop']['struct'],
+                           self.MAGIC['request'],
+                           self.COMMANDS['noop']['command'],
+                           0, 0, 0, 0, 0, 0, 0)
+        self._send(data)
+
+        (magic, opcode, keylen, extlen, datatype, status, bodylen, opaque,
+         cas, extra_content) = self._get_response()
+
+        logger.debug('Value Length: %d. Body length: %d. Data type: %d',
+                     extlen, bodylen, datatype)
+
+        if status != self.STATUS['success']:
+            logger.debug('NOOP failed (status is %d). Message: %s' % (status, extra_content))
+
+        return int(status)
+
     def get_multi(self, keys):
         """
         Get multiple keys from server.
@@ -489,7 +533,7 @@ class Protocol(threading.local):
             elif status == self.STATUS['server_disconnected']:
                 break
             elif status != self.STATUS['key_not_found']:
-                raise MemcachedException('Code: %d Message: %s' % (status, extra_content))
+                raise MemcachedException('Code: %d Message: %s' % (status, extra_content), status)
 
         return d
 
@@ -536,7 +580,7 @@ class Protocol(threading.local):
                 return False
             elif status == self.STATUS['server_disconnected']:
                 return False
-            raise MemcachedException('Code: %d Message: %s' % (status, extra_content))
+            raise MemcachedException('Code: %d Message: %s' % (status, extra_content), status)
 
         return True
 
@@ -724,7 +768,7 @@ class Protocol(threading.local):
          cas, extra_content) = self._get_response()
 
         if status not in (self.STATUS['success'], self.STATUS['server_disconnected']):
-            raise MemcachedException('Code: %d Message: %s' % (status, extra_content))
+            raise MemcachedException('Code: %d Message: %s' % (status, extra_content), status)
         if status == self.STATUS['server_disconnected']:
             return 0
 
@@ -789,7 +833,7 @@ class Protocol(threading.local):
         if status == self.STATUS['server_disconnected']:
             return False
         if status != self.STATUS['success'] and status not in (self.STATUS['key_not_found'], self.STATUS['key_exists']):
-            raise MemcachedException('Code: %d message: %s' % (status, extra_content))
+            raise MemcachedException('Code: %d message: %s' % (status, extra_content), status)
 
         logger.debug('Key deleted %s', key)
         return status != self.STATUS['key_exists']
@@ -857,7 +901,7 @@ class Protocol(threading.local):
          cas, extra_content) = self._get_response()
 
         if status not in (self.STATUS['success'], self.STATUS['server_disconnected']):
-            raise MemcachedException('Code: %d message: %s' % (status, extra_content))
+            raise MemcachedException('Code: %d message: %s' % (status, extra_content), status)
 
         logger.debug('Memcached flushed')
         return True
